import pathlib
from typing import List

from ladybug_geometry.geometry3d import Face3D, Polyface3D, Point3D, Vector3D
from honeybee.model import Model, Shade, Room

from .templates import SPACE_TEMPLATE, SHADE_TEMPLATE, ADJ_BLDG_TEMPLATE


def _opening_to_ies(
        parent_geo: Face3D, opening_geos: List[Face3D], opening_type: int = 0) -> str:
    """Translate an opening to gem format.

    Args:
        parent_geo: Geometry of the parent object.
        opening_geos: A list of Face3D geometries for openings.
        opening_type: An integer between 0-2. 0 for apertures, 1 for doors and 2 for
            holes.

    Returns:
        A formatted string for the opening.

    """
    if parent_geo.plane.n.z in (1, -1):
        origin, flip = parent_geo.upper_right_corner, True
    else:
        origin, flip = parent_geo.lower_left_corner, False
    openings = []
    for opening in opening_geos:
        verts_2d = parent_geo.polygon_in_face(opening, origin, flip)
        openings.append('{} {}'.format(len(verts_2d), opening_type))
        openings.append(
            '\n'.join('   {:.6f}    {:.6f}'.format(v.x, v.y) for v in verts_2d)
        )
    return '\n'.join(openings)


def _vertices_to_ies(vertices: List[Point3D]) -> str:
    """Get a string for vertices in GEM format."""
    vertices = '\n'.join(
        '   {:.6f}    {:.6f}    {:.6f}'.format(v.x, v.y, v.z)
        for v in vertices
    )
    return vertices


def _shade_geometry_to_ies(geometry: Polyface3D, name: str, is_detached=True):
    unique_vertices = geometry.vertices
    vertices = _vertices_to_ies(unique_vertices)
    index = [str(i + 1) for i in range(len(unique_vertices))]
    faces = []
    for face_i in geometry.face_indices:
        index = [str(v + 1) for v in face_i[0]]
        face = '%d %s \n0' % (len(index), ' '.join(index))
        faces.append(face)

    template = ADJ_BLDG_TEMPLATE if is_detached else SHADE_TEMPLATE
    return template.format(
        name=name,
        vertices_count=len(unique_vertices),
        vertices=vertices,
        faces='\n'.join(faces), face_count=len(geometry.faces)
    )


def _shade_group_to_ies(shades: List[Shade]) -> str:
    """Convert a group of shades into a GEM string.

    The files in the shade group should create a closed volume. The translator uses
    the name of the first shade as the name of the group.
    """
    group_geometry = Polyface3D.from_faces(
        [shade.geometry for shade in shades], tolerance=0.001
    )
    first_shade = shades[0]
    return _shade_geometry_to_ies(
        group_geometry, first_shade.display_name, first_shade.is_detached
    )



def _shade_to_ies(shade: Shade, thickness: float = 0.1) -> str:
    """Convert a single Honeybee Shade to a GEM string.

    Args:
        shade: A Shade face.
        thickness:The thickness of the shade face in meters. IES doesn't consider the
            effect of shades with no thickness in SunCalc. This function extrudes the
            geometry to create a closed volume for the shade. Default: 0.1

    Returns:
        A formatted string that represents this shade in GEM format.

    """
    geometry = shade.geometry
    move_vector = geometry.normal.reverse().normalize() * thickness / 2
    base_geo = geometry.move(move_vector)
    shade_geo = Polyface3D.from_offset_face(base_geo, thickness)
    return _shade_geometry_to_ies(
        shade_geo, shade.display_name, is_detached=shade.is_detached
    )


def shades_to_ies(shades: List[Shade], thickness: float = 0.1) -> str:
    """Convert a list of Shades to a GEM string.

    Args:
        shades: A list of Shade faces.
        thickness:The thickness of the shade face in meters. This value will be used to
            extrude shades with no group id. IES doesn't consider the effect of shades
            with no thickness in SunCalc. This function extrudes the geometry to create
            a closed volume for the shade. Default: 0.1

    Returns:
        A formatted string that represents this shade in GEM format.

    """
    shade_groups = {}
    no_groups = []
    for shade in shades:
        try:
            group_id = shade.user_data['__group_id__']
        except (TypeError, KeyError):
            no_groups.append(shade)
            continue
        else:
            if group_id not in shade_groups:
                shade_groups[group_id] = [shade]
            else:
                shade_groups[group_id].append(shade)

    single_shades = '\n'.join([_shade_to_ies(shade) for shade in no_groups])
    group_shades = '\n'.join(
        [_shade_group_to_ies(shades) for shades in shade_groups.values()]
        )

    return '\n'.join((single_shades, group_shades))


def room_to_ies(room: Room) -> str:
    """Convert a Honeybee Shade to a GEM string.

    Args:
        room: A Honeybee Room.

    Returns:
        A formatted string that represents this room in GEM format.

    """
    unique_vertices = room.geometry.vertices
    vertices = _vertices_to_ies(unique_vertices)

    faces = []
    for face_i, face in zip(room.geometry.face_indices, room.faces):
        index = [str(v + 1) for v in face_i[0]]
        face_str = '%d %s \n' % (len(index), ' '.join(index))
        open_count, openings, fg = 0, [], face.geometry
        if fg.has_holes:
            sub_faces = [Face3D(hole, fg.plane) for hole in fg.holes]
            openings.append(_opening_to_ies(fg, sub_faces, 2))
            open_count += len(sub_faces)
        if len(face.apertures) != 0:
            sub_faces = [ap.geometry for ap in face.apertures]
            openings.append(_opening_to_ies(fg, sub_faces, 0))
            open_count += len(sub_faces)
        if len(face.doors) != 0:
            sub_faces = [dr.geometry for dr in face.doors]
            openings.append(_opening_to_ies(fg, sub_faces, 1))
            open_count += len(sub_faces)
        open_str = '\n' + '\n'.join(openings) if len(openings) != 0 else ''
        faces.append('%s%d%s' % (face_str, open_count, open_str))

    space = SPACE_TEMPLATE.format(
        space_name=room.display_name, vertices_count=len(unique_vertices),
        face_count=len(room.faces), vertices=vertices, faces='\n'.join(faces)
    )

    # collect all the shades from room
    shades = [shade for shade in room.shades]
    for face in room.faces:
        for aperture in face.apertures:
            for shade in aperture.shades:
                shades.append(shade)
        for door in face.doors:
            for shade in door.shades:
                shades.append(shade)
    if shades:
        formatted_shades = shades_to_ies(shades=shades)
        return '\n'.join((space, formatted_shades))
    else:
        return space


def model_to_ies(
        model: Model, folder: str = '.', name: str = None) -> pathlib.Path:
    """Export a honeybee model to an IES GEM file.

    Args:
        model: A honeybee model.
        folder: Path to target folder to export the file. Default is current folder.
        name: An optional name for exported file. By default the name of the model will
            be used.

    Returns:
        Path to exported GEM file.
    """
    # ensure model is in metrics
    model.convert_to_units(units='Meters')

    header = 'COM GEM data file exported by Pollination\n' \
        'ANT\n'
    rooms_data = [room_to_ies(room) for room in model.rooms]
    context_shades = shades_to_ies(model.shades)

    # write to GEM
    name = name or model.display_name
    if not name.lower().endswith('.gem'):
        name = f'{name}.gem'
    out_folder = pathlib.Path(folder)
    out_folder.mkdir(parents=True, exist_ok=True)
    out_file = out_folder.joinpath(name)
    with out_file.open('w') as outf:
        outf.write(header)
        outf.write('\n'.join(rooms_data) + '\n')
        outf.write(context_shades)

    return out_file
